@@ -1,10 +1,11 @@
+import re
 import sys
 import json
+import types
 import shutil
 import typing
 import logging
 import pathlib
-import argparse
 import itertools
 import collections
 import collections.abc
@@ -13,7 +14,7 @@ import urllib.request
 
 import attr
 import csvw
-from csvw.metadata import TableGroup, Table, Column, Link, Schema, is_url
+from csvw.metadata import TableGroup, Table, Column, Link, Schema, is_url, URITemplate
 from csvw import datatypes
 from csvw.dsv import iterrows
 from clldutils.path import git_describe, walk
@@ -21,7 +22,7 @@ from clldutils.misc import log_or_raise, lazyproperty
 from clldutils import jsonlib
 
 from pycldf.sources import Sources
-from pycldf.util import pkg_path, resolve_slices, DictTuple, sanitize_url
+from pycldf.util import pkg_path, resolve_slices, DictTuple, sanitize_url, iter_uritemplates
 from pycldf.terms import term_uri, Terms, TERMS, get_column_names, URL as TERMS_URL
 from pycldf.validators import VALIDATORS
 from pycldf import orm
@@ -32,6 +33,9 @@ __all__ = [
 
 MD_SUFFIX = '-metadata.json'
 ORM_CLASSES = {cls.component_name(): cls for cls in orm.Object.__subclasses__()}
+TableType = typing.Union[str, Table]
+ColType = typing.Union[str, Column]
+PathType = typing.Union[str, pathlib.Path]
 
 
 class SchemaError(KeyError):
@@ -39,7 +43,7 @@ class SchemaError(KeyError):
 
 
 @attr.s
-class Module(object):
+class Module:
     """
     Class representing a CLDF Module.
 
@@ -50,13 +54,13 @@ class Module(object):
     cls = attr.ib(default=None)
 
     @property
-    def id(self):
+    def id(self) -> str:
         """
         The local part of the term URI is interpreted as Module identifier.
         """
         return self.uri.split('#')[1]
 
-    def match(self, thing):
+    def match(self, thing) -> bool:
         if isinstance(thing, TableGroup):
             return thing.common_props.get('dc:conformsTo') == term_uri(self.id)
         if hasattr(thing, 'name'):
@@ -67,7 +71,7 @@ class Module(object):
 _modules = []
 
 
-def get_modules():
+def get_modules() -> typing.List[Module]:
     """
     We read supported CLDF modules from the default metadata files distributed with `pycldf`.
     """
@@ -88,7 +92,7 @@ def get_modules():
     return _modules
 
 
-def make_column(spec):
+def make_column(spec: typing.Union[str, dict, Column]) -> Column:
     if isinstance(spec, str):
         if spec in TERMS.by_uri:
             return TERMS.by_uri[spec].to_column()
@@ -100,7 +104,7 @@ def make_column(spec):
     raise TypeError(spec)
 
 
-class GitRepository(object):
+class GitRepository:
     def __init__(self, url, clone=None, version=None, **dc):
         # We remove credentials from the URL immediately to make sure this isn't leaked into
         # CLDF metadata. Such credentials might be present in URLs read via gitpython from
@@ -123,15 +127,15 @@ class GitRepository(object):
         return res
 
 
-class Dataset(object):
+class Dataset:
     """
     API to access a CLDF dataset.
     """
 
     def __init__(self, tablegroup: csvw.TableGroup):
         """
-        A :class:`~pycldf.dataset.Dataset` is initialized passing a TableGroup. For convenience \
-        methods to get such a TableGroup, see the factory methods
+        A :class:`~pycldf.dataset.Dataset` is initialized passing a `TableGroup`. For convenience \
+        methods to get such a `TableGroup` instance, see the factory methods
 
         - :meth:`~pycldf.dataset.Dataset.in_dir`
         - :meth:`~pycldf.dataset.Dataset.from_metadata`
@@ -147,7 +151,7 @@ class Dataset(object):
     # Factory methods to create `Dataset` instances.
     #
     @classmethod
-    def in_dir(cls, d: typing.Union[str, pathlib.Path], empty_tables=False):
+    def in_dir(cls, d: PathType, empty_tables: bool = False) -> 'Dataset':
         """
         Create a :class:`~pycldf.dataset.Dataset` in a (possibly empty or even non-existing) \
         directory.
@@ -166,7 +170,7 @@ class Dataset(object):
         return res
 
     @classmethod
-    def from_metadata(cls, fname: typing.Union[str, pathlib.Path]):
+    def from_metadata(cls, fname: PathType) -> 'Dataset':
         """
         Initialize a :class:`~pycldf.dataset.Dataset` with the metadata found at `fname`.
 
@@ -205,7 +209,7 @@ class Dataset(object):
         return cls(tablegroup)
 
     @classmethod
-    def from_data(cls, fname):
+    def from_data(cls, fname: PathType) -> 'Dataset':
         """
         Initialize a :class:`~pycldf.dataset.Dataset` from a single CLDF data file.
 
@@ -251,7 +255,7 @@ class Dataset(object):
         return self.properties['dc:conformsTo'].split('#')[1]
 
     @property
-    def version(self):
+    def version(self) -> str:
         return self.properties['dc:conformsTo'].split('/')[3]
 
     def __repr__(self):
@@ -423,7 +427,7 @@ class Dataset(object):
         except SchemaError:
             return default
 
-    def get_foreign_key_reference(self, table, column) \
+    def get_foreign_key_reference(self, table: TableType, column: ColType) \
             -> typing.Union[typing.Tuple[csvw.Table, csvw.Column], None]:
         """
         Retrieve the reference of a foreign key constraint for the specified column.
@@ -441,7 +445,7 @@ class Dataset(object):
                     self[fk.reference.resource, fk.reference.columnReference[0]]
 
     @property
-    def column_names(self) -> argparse.Namespace:
+    def column_names(self) -> types.SimpleNamespace:
         """
         In-direction layer, mapping ontology terms to local column names (or `None`).
 
@@ -449,18 +453,18 @@ class Dataset(object):
         schema may have changed). So when accessing a dataset for reading only, calling code
         should use `readonly_column_names`.
 
-        :return: an `argparse.Namespace` object, with attributes `<object>s` for each component \
+        :return: an `types.SimpleNamespace` object, with attributes `<object>s` for each component \
         `<Object>Table` defined in the ontology. Each such attribute evaluates to `None` if the \
-        dataset does not contain the component. Otherwise, it's an `argparse.Namespace` object \
+        dataset does not contain the component. Otherwise, it's an `types.SimpleNamespace` object \
         mapping each property defined in the ontology to `None` - if no such column is specified \
         in the component - and the local column name if it is.
         """
         return get_column_names(self)
 
     @lazyproperty
-    def readonly_column_names(self) -> argparse.Namespace:
+    def readonly_column_names(self) -> types.SimpleNamespace:
         """
-        :return: `argparse.Namespace` with component names as attributes.
+        :return: `types.SimpleNamespace` with component names as attributes.
         """
         return get_column_names(self, use_component_names=True, with_multiplicity=True)
 
@@ -512,7 +516,7 @@ class Dataset(object):
                 kw.pop('primaryKey'))
         return t
 
-    def remove_table(self, table):
+    def remove_table(self, table: TableType):
         """
         Removes the table specified by `table` from the dataset.
         """
@@ -567,7 +571,7 @@ class Dataset(object):
         self.auto_constraints(component)
         return component
 
-    def add_columns(self, table, *cols) -> None:
+    def add_columns(self, table: TableType, *cols) -> None:
         """
         Add columns specified by `cols` to the table specified by `table`.
         """
@@ -584,11 +588,11 @@ class Dataset(object):
             table.tableSchema.columns.append(col)
         self.auto_constraints()
 
-    def remove_columns(self, table, *cols):
+    def remove_columns(self, table: TableType, *cols):
         """
         Remove `cols` from `table`'s schema.
 
-        Note: Foreign keys pointing to any of the removed columns are removed as well.
+        .. note:: Foreign keys pointing to any of the removed columns are removed as well.
         """
         table = self[table]
         cols = [str(self[table, col]) for col in cols]
@@ -607,11 +611,52 @@ class Dataset(object):
 
         table.tableSchema.columns = [c for c in table.tableSchema.columns if str(c) not in cols]
 
-    def add_foreign_key(self, foreign_t, foreign_c, primary_t, primary_c=None):
+    def rename_column(self, table: TableType, col: ColType, name: str):
+        """
+        Assign a new `name` to an existing column, cascading this change to foreign keys.
+
+        This functionality can be used to change the names of columns added automatically by
+        :meth:`Dataset.add_component`
+        """
+        table = self[table]
+        col = self[table, col]
+
+        for obj, prop, tmpl in iter_uritemplates(table):
+            if col.name in tmpl.variable_names:
+                old = str(tmpl)
+                new = re.sub(
+                    r'{([^}]+)}',
+                    lambda m: '{' + re.sub(re.escape(col.name), name, m.groups()[0]) + '}',
+                    old)
+                if old != new:
+                    setattr(obj, prop, URITemplate(new))
+
+        if col.name in table.tableSchema.primaryKey:
+            table.tableSchema.primaryKey = [
+                name if n == col.name else n for n in table.tableSchema.primaryKey]
+
+        for t in self.tables:
+            for fk in t.tableSchema.foreignKeys:
+                if fk.reference.resource == table.url and col.name in fk.reference.columnReference:
+                    fk.reference.columnReference = [
+                        name if n == col.name else n for n in fk.reference.columnReference]
+
+                if t.url == table.url and col.name in fk.columnReference:
+                    # We also need to check the columnReference
+                    fk.columnReference = [name if n == col.name else n for n in fk.columnReference]
+
+        col.name = name
+
+    def add_foreign_key(
+            self,
+            foreign_t: TableType,
+            foreign_c: ColType,
+            primary_t: TableType,
+            primary_c: typing.Optional[ColType] = None):
         """
         Add a foreign key constraint.
 
-        Note: Composite keys are not supported yet.
+        ..note:: Composite keys are not supported yet.
 
         :param foreign_t: Table reference for the linking table.
         :param foreign_c: Column reference for the link.
@@ -700,7 +745,7 @@ class Dataset(object):
     #
     # Methods to read data
     #
-    def iter_rows(self, table, *cols) -> typing.Iterator[dict]:
+    def iter_rows(self, table: TableType, *cols) -> typing.Iterator[dict]:
         """
         Iterate rows in a table, resolving CLDF property names to local column names.
 
@@ -716,7 +761,7 @@ class Dataset(object):
                 item[v] = item[k]
             yield item
 
-    def get_row(self, table, id_) -> dict:
+    def get_row(self, table: TableType, id_) -> dict:
         """
         Retrieve a row specified by table and CLDF id.
 
@@ -728,7 +773,7 @@ class Dataset(object):
                 return row
         raise ValueError(id_)  # pragma: no cover
 
-    def get_row_url(self, table, row) -> typing.Union[str, None]:
+    def get_row_url(self, table: TableType, row) -> typing.Union[str, None]:
         """
         Get a URL associated with a row. Tables can specify associated row URLs by
 
@@ -756,7 +801,7 @@ class Dataset(object):
         if id_col.valueUrl:
             return id_col.valueUrl.expand(**row)
 
-    def objects(self, table, cls=None) -> DictTuple:
+    def objects(self, table: str, cls: typing.Optional[typing.Type] = None) -> DictTuple:
         """
         Read data of a CLDF component as :class:`pycldf.orm.Object` instances.
 
@@ -787,7 +832,8 @@ class Dataset(object):
     #
     # Methods for writing (meta)data to files:
     #
-    def write_metadata(self, fname=None):
+    def write_metadata(
+            self, fname: typing.Optional[typing.Union[str, pathlib.Path]] = None) -> pathlib.Path:
         """
         Write the CLDF metadata to a JSON file.
 
@@ -931,6 +977,13 @@ class Dataset(object):
                                 table_uri, uri, cardinality), log=log)
 
         for table in self.tables:
+            vars = set(col.name for col in table.tableSchema.columns)
+            for obj, prop, tmpl in iter_uritemplates(table):
+                if not {n for n in tmpl.variable_names if not n.startswith('_')}.issubset(vars):
+                    if log:
+                        log.warning('Unknown variables in URI template: {}:{}:{}'.format(
+                            obj, prop, tmpl))
+
             type_uri = table.common_props.get('dc:conformsTo')
             if type_uri:
                 try:
@@ -953,7 +1006,7 @@ class Dataset(object):
             # FIXME: check whether table.common_props['dc:conformsTo'] is in validators!
             validators_, propertyUrls, colnames = [], set(), set()
             for col in table.tableSchema.columns:
-                if col.header in colnames:
+                if col.header in colnames:  # pragma: no cover
                     log_or_raise(
                         'Duplicate column name in table schema: {} {}'.format(
                             table.url, col.header),
@@ -963,7 +1016,7 @@ class Dataset(object):
                     col_uri = col.propertyUrl.uri
                     try:
                         terms.is_cldf_uri(col_uri)
-                        if col_uri in propertyUrls:
+                        if col_uri in propertyUrls:  # pragma: no cover
                             log_or_raise(
                                 'Duplicate CLDF property in table schema: {} {}'.format(
                                     table.url, col_uri),
@@ -1025,17 +1078,27 @@ class Dataset(object):
 
 
 class Generic(Dataset):
+    """
+    Generic datasets have no primary table.
+
+    .. seealso:: `<https://github.com/cldf/cldf/tree/master/modules/Generic>`_
+    """
     @property
     def primary_table(self):
         return None
 
 
 class Wordlist(Dataset):
+    """
+    Wordlists have support for segment slice notation.
+
+    .. seealso:: `<https://github.com/cldf/cldf/tree/master/modules/Wordlist>`_
+    """
     @property
     def primary_table(self):
         return 'FormTable'
 
-    def get_segments(self, row, table='FormTable'):
+    def get_segments(self, row, table='FormTable') -> typing.List[str]:
         col = self[table].get_column("http://cldf.clld.org/v1.0/terms.rdf#segments")
         sounds = row[col.name]
         if isinstance(sounds, str):
@@ -1043,13 +1106,12 @@ class Wordlist(Dataset):
             sounds = [sounds]
         return list(itertools.chain(*[s.split() for s in sounds]))
 
-    def get_subsequence(self, cognate, form=None):
+    def get_subsequence(self, cognate: dict, form=None) -> typing.List[str]:
         """
         Compute the subsequence of the morphemes of a form which is specified in a partial
         cognate assignment.
 
-        :param partial_cognate:
-        :return:
+        :param cognate: A `dict` holding the data of a row from a `CognateTable`.
         """
         return resolve_slices(
             cognate,
@@ -1119,4 +1181,8 @@ def iter_datasets(d: pathlib.Path) -> typing.Iterator[Dataset]:
     """
     for p in walk(d, mode='files'):
         if sniff(p):
-            yield Dataset.from_metadata(p)
+            try:
+                yield Dataset.from_metadata(p)
+            except ValueError as e:
+                logging.getLogger(__name__).warning(
+                    "Reading {} failed: {}".format(p, e))
